#ifndef __LIBGRAVIS_T_IMAGE_ALGORITHM_H__
#define __LIBGRAVIS_T_IMAGE_ALGORITHM_H__
/******************************************************************************
**        Title: tImageAlgorithm.h
**  Description: Collection of standalone image algorithms.
**
**       Author: Jean-Sebastien Pierrard, 2005
**               Brian Schroeder 2006
**               Computer Science Department, University Basel (CH)
**
******************************************************************************/

#include "tImage/convolution.h"
#include "tImage/normalization.h"
#include "tImage/operators.h"

#endif
