#include <signal.h>
#include <cuda_runtime.h>
#include "src/acc/cuda/cuda_settings.h"
#include "src/acc/acc_backprojector.h"
#include "src/acc/acc_projector.h"

#include "src/acc/acc_backprojector_impl.h"
